/*
TCP Client/Server Performance Test Application
Copyright (C) 2005-2014, James Gibbons
Released under the GNU Lesser General Public License version 2.1
See http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html
*/
#include "qt_tcp_clientserverapp.h"
#include "ui_qt_tcp_clientserverapp.h"
#include <QDateTime>
#include <QTextStream>

//---------------------------------------------------------------------------
// this is the Ethernet limit for UDP data frames
// (actually 1480 but it limits to 1472 due to overhead)
static const int UDP_DATA_SIZE=1472;
// optional values to try
//static const int UDP_DATA_SIZE=512;
//static const int UDP_DATA_SIZE=32;
//---------------------------------------------------------------------------

//---------------------------------------------------------------------
Qt_TCP_ClientServerApp::Qt_TCP_ClientServerApp(QWidget *parent)
	: QMainWindow(parent),
		ui(new Ui::Qt_TCP_ClientServerAppClass)
{
	ui->setupUi(this);

	IOBuffer=NULL;

	tHRT.HighResReset();
	rHRT.HighResReset();

	connect(&runTimer,SIGNAL(timeout()),this,SLOT(runTimerTick()));

	theServer=new QTcpServer(this);
	connect(theServer,SIGNAL(newConnection()),this,SLOT(serverConnect()));

	serverSocket=NULL;

	clientSocket=new QTcpSocket(this);
	connect(clientSocket,SIGNAL(connected()),this,SLOT(clientConnect()));
	connect(clientSocket,SIGNAL(readyRead()),this,SLOT(clientRead()));
	connect(clientSocket,SIGNAL(disconnected()),this,SLOT(clientDisconnect()));
	connect(clientSocket,SIGNAL(error(QAbstractSocket::SocketError))
					,this,SLOT(clientError(QAbstractSocket::SocketError)));

	udpSocket=new QUdpSocket(this);
	connect(udpSocket,SIGNAL(readyRead()),this,SLOT(udpRead()));
	connect(udpSocket,SIGNAL(error(QAbstractSocket::SocketError))
					,this,SLOT(udpError(QAbstractSocket::SocketError)));

	ui->recMode->clear();
	ui->recMode->insertItem(0,"Binary Stream");
	ui->recMode->insertItem(1,"Reserved");
	ui->recMode->insertItem(2,"UDP");
	ui->recMode->setCurrentIndex(0);
	ui->ipAddress->setText("127.0.0.1");
	NetStatus=NET_RESET;
	SetButtons(true);
}

Qt_TCP_ClientServerApp::~Qt_TCP_ClientServerApp()
{
	on_stopButton_clicked();
	delete ui;
}

void Qt_TCP_ClientServerApp::SetButtons(bool bState)
{
	ui->sendData->setCheckState(Qt::Unchecked);
	ui->recMode->setEnabled(bState);
	ui->serverButton->setEnabled(bState);
	ui->clientButton->setEnabled(bState);
	ui->stopButton->setEnabled(!bState);
}

void Qt_TCP_ClientServerApp::AddMessage(QString s)
{
	QDateTime dt=QDateTime::currentDateTime();
	ui->messageDisplay->insertPlainText(dt.toString("MM-dd-yy hh:mm:ss.zzz ")+s+"\n");
	ui->messageDisplay->moveCursor(QTextCursor::End);
}

void Qt_TCP_ClientServerApp::on_stopButton_clicked()
{
	runTimer.stop();
	ui->sendData->setCheckState(Qt::Unchecked);
	ui->sendData->setEnabled(true);
	switch (NetStatus)
	{
		case NET_RESET:
			SetButtons(true);
			break;
		case NET_SOCK_SERVER:
			theServer->close();
			if (serverSocket!=NULL)
			{
				serverSocket->close();
				serverSocket=NULL;
			}
			udpSocket->close();
			AddMessage("Server socket is closed...");
			SetButtons(true);
			NetStatus=NET_RESET;
			ReleaseIOBuffer();
			break;
		case NET_SOCK_CLIENT:
			switch (ui->recMode->currentIndex())
			{
				case 0:
				case 1:
					clientSocket->close();
					AddMessage("Client socket is closed...");
					SetButtons(true);
					NetStatus=NET_RESET;
					ReleaseIOBuffer();
					break;
				case 2:
					udpSocket->close();
					AddMessage("UDP client is closed...");
					SetButtons(true);
					NetStatus=NET_RESET;
					ReleaseIOBuffer();
					// delete the socket to break the binding
					delete udpSocket;
					udpSocket=NULL;
					udpSocket=new QUdpSocket(this);
					connect(udpSocket,SIGNAL(readyRead()),this,SLOT(udpRead()));
					connect(udpSocket,SIGNAL(error(QAbstractSocket::SocketError))
									,this,SLOT(udpError(QAbstractSocket::SocketError)));
					break;
			}
			break;
	}
}

void Qt_TCP_ClientServerApp::on_serverButton_clicked()
{
	ui->messageDisplay->clear();
	SetButtons(false);
	AddMessage("Socket server is entering listen mode...");
	CheckBuf=0;
	NSend=TSend=0;
	NRec=TRec=0;
	tHRT.HighResReset();
	rHRT.HighResReset();
	NewIOBuffer();
	serverSocket=NULL;
	NetStatus=NET_SOCK_SERVER;
	switch (ui->recMode->currentIndex())
	{
		case 0:
		case 1:
			if (!theServer->listen(QHostAddress::Any,1667))
			{
				AddMessage("Socket server did not start: "+theServer->errorString());
				on_stopButton_clicked();
				break;
			}
			runTimer.start(50);
			break;
		case 2:
			if (!udpSocket->bind(1667))
			{
				AddMessage("UDP socket did not bind: "+udpSocket->errorString());
				on_stopButton_clicked();
				break;
			}
			ui->sendData->setEnabled(false);
			runTimer.start(50);
			break;
	}
}

void Qt_TCP_ClientServerApp::on_clientButton_clicked()
{
	ui->messageDisplay->clear();
	SetButtons(false);
	ui->sendData->setEnabled(true);
	CheckBuf=0;
	NSend=TSend=0;
	NRec=TRec=0;
	tHRT.HighResReset();
	rHRT.HighResReset();
	NewIOBuffer();
	bool cFlag=false;

	switch (ui->recMode->currentIndex())
	{
		case 0:
		case 1:
			clientSocket->close();
			AddMessage("Socket client is attempting connection to "+ui->ipAddress->text());
			qApp->processEvents();

			clientSocket->connectToHost(ui->ipAddress->text(),1667);

			cFlag=clientSocket->waitForConnected(2000);

			if (!cFlag)
			{
				QHostInfo info = QHostInfo::fromName(ui->ipAddress->text());

				for (int i=0;i<info.addresses().count();i++)
				{
					clientSocket->connectToHost(info.addresses().at(i),1667);
					cFlag=clientSocket->waitForConnected(2000);
					if (cFlag)
						break;
				}
			}

			if (!cFlag)
			{
				AddMessage("Connection timeout abort...");
				on_stopButton_clicked();
			}
			else
			{
				NetStatus=NET_SOCK_CLIENT;
				runTimer.start(50);
			}
			break;
		case 2:
			udpSocket->close();
			AddMessage("UDP socket is ready to send to "+ui->ipAddress->text());
			NetStatus=NET_SOCK_CLIENT;
			runTimer.start(50);
			break;
	}
}

void Qt_TCP_ClientServerApp::on_sendData_clicked(bool c)
{
	tHRT.HighResReset();
	if (runTimer.isActive())
	{
		if (c)
		{
			switch (ui->recMode->currentIndex())
			{
				case 0:
				case 1:
					// run TCP at idle rate
					runTimer.setInterval(0);
					break;
				case 2:
					// run UDP at 1 msec timing
					runTimer.setInterval(1);
					break;
			}
		}
		else
			runTimer.setInterval(50);
	}
}

void Qt_TCP_ClientServerApp::runTimerTick(void)
{
	switch (NetStatus)
	{
		case NET_RESET:
			break;
		case NET_SOCK_CLIENT:
			if (ui->sendData->isChecked())
			{
				switch (ui->recMode->currentIndex())
				{
					case 0:
					case 1:
						// send record on every tick if buffer is empty
						if (clientSocket->bytesToWrite()<SizeIOBuffer)
						{
							qint64 Sent=clientSocket->write(IOBuffer,SizeIOBuffer);
							if (Sent>0) NSend+=Sent;
						}
						break;
					case 2:
						{
							const int bsize=UDP_DATA_SIZE;
							UDPString.replace(0,1,UDPSig);
							UDPSig++;
							if (UDPSig>'Z') UDPSig='A';
							QByteArray barray(UDPString.toLatin1());
							QHostAddress hadr(ui->ipAddress->text());
							qint64 NS=udpSocket->writeDatagram(barray.data(),bsize,hadr,1667);
							if (NS>0)
								NSend+=NS;
						}
						break;
				}
			}
			break;
		case NET_SOCK_SERVER:
			if (ui->sendData->isChecked())
			{
				switch (ui->recMode->currentIndex())
				{
					case 0:
					case 1:
						if (serverSocket==NULL) break;
						// send record on every tick if buffer is empty
						if (serverSocket->bytesToWrite()<SizeIOBuffer)
						{
							int Sent=serverSocket->write(IOBuffer,SizeIOBuffer);
							if (Sent>0) NSend+=Sent;
						}
						break;
					case 2:
						// UDP server doesn't send
						break;
				}
			}
			break;
	}

	double t=tHRT.HighResTime();
	if (t>=5.0)
	{
		tHRT.HighResReset();
		int NS=NSend;
		NSend=0;
		TSend+=NS;

		if (NS>0)
		{
			QString s;
			QTextStream ts(&s);
			ts << TSend << " total bytes sent, "
				 << int(NS/t) << " bytes per second...";
			AddMessage(s);
		}
	}

	t=rHRT.HighResTime();
	if (t>=5.0)
	{
		rHRT.HighResReset();
		int NR=NRec;
		NRec=0;
		TRec+=NR;

		if (NR>0)
		{
			QString s;
			QTextStream ts(&s);
			ts << TRec << " total bytes received, "
				 << int(NR/t) << " bytes per second...";
			AddMessage(s);
		}

		if (UDPSigErrors>0)
		{
			QString s;
			QTextStream ts(&s);
			ts << "Sequence errors: " << UDPSigErrors;
			AddMessage(s);
			UDPSigErrors=0;
		}
	}
}

void Qt_TCP_ClientServerApp::NewIOBuffer(void)
{
	ReleaseIOBuffer();
	IOBuffer=new char [BUF_SIZE];
	for (int i=0;i<BUF_SIZE;i++) IOBuffer[i]=(char)i;
	SizeIOBuffer=BUF_SIZE;

	const int bsize=UDP_DATA_SIZE;
	const char minc='A';
	const char maxc='Z';
	char c=minc;

	UDPString="";

	for (int i=0;i<bsize;i++)
	{
		UDPString=UDPString+c;
		c++;
		if (c>maxc) c=minc;
	}

	UDPSig=UDPCheckSig='A';
	UDPSigErrors=0;
}

void Qt_TCP_ClientServerApp::ReleaseIOBuffer(void)
{
	if (IOBuffer!=NULL)
	{
		SizeIOBuffer=0;
		delete[] IOBuffer;
		IOBuffer=NULL;
	}
	UDPString.clear();
}

void Qt_TCP_ClientServerApp::serverConnect(void)
{
	QTcpSocket *newSocket=theServer->nextPendingConnection();
	if (newSocket!=NULL)
	{
		QString s;
		QTextStream ts(&s);
		ts << "Client " << newSocket->peerAddress().toString()
			 << " is connecting on port " << newSocket->peerPort();
		AddMessage(s);
		if (serverSocket==NULL)
		{
			serverSocket=newSocket;
			connect(serverSocket,SIGNAL(readyRead()),this,SLOT(serverRead()));
			connect(serverSocket,SIGNAL(disconnected()),this,SLOT(serverDisconnect()));
			connect(serverSocket,SIGNAL(error(QAbstractSocket::SocketError))
							,this,SLOT(serverError(QAbstractSocket::SocketError)));
		}
		else
		{
			newSocket->close();
			AddMessage("Server is active, new connection rejected...");
		}
	}
}

void Qt_TCP_ClientServerApp::serverRead(void)
{
	if (serverSocket==NULL) return;

	qint64 nbytes;

	switch (ui->recMode->currentIndex())
	{
		case 0:
		case 1:
			nbytes=serverSocket->bytesAvailable();
			if (nbytes>0)
			{
				char *c=new char[nbytes];
				qint64 nb=serverSocket->read(c,nbytes);
				NRec+=nb;
				delete[] c;
			}
			break;
		case 2:
			break;
	}
}

void Qt_TCP_ClientServerApp::serverDisconnect(void)
{
	AddMessage("Client disconnected...");
	serverSocket=NULL;
}

void Qt_TCP_ClientServerApp::serverError(QAbstractSocket::SocketError socketError)
{
	switch (socketError)
	{
		case QAbstractSocket::RemoteHostClosedError:
			AddMessage("Remote Host Closed error: "+serverSocket->errorString());
			break;
		case QAbstractSocket::HostNotFoundError:
			AddMessage("Host Not Found error: "+serverSocket->errorString());
			break;
		case QAbstractSocket::ConnectionRefusedError:
			AddMessage("Connection Refused error: "+serverSocket->errorString());
			break;
		default:
			AddMessage("Server socket error: "+serverSocket->errorString());
			break;
	}
}

void Qt_TCP_ClientServerApp::clientConnect(void)
{
	QString s;
	QTextStream ts(&s);
	ts << "Client connected to host " << clientSocket->peerAddress().toString()
		 << " on port " << clientSocket->peerPort();
	AddMessage(s);
}

void Qt_TCP_ClientServerApp::clientRead(void)
{
	qint64 nbytes;

	switch (ui->recMode->currentIndex())
	{
		case 0:
		case 1:
			nbytes=clientSocket->bytesAvailable();
			if (nbytes>0)
			{
				char *c=new char[nbytes];
				qint64 nb=clientSocket->read(c,nbytes);
				NRec+=nb;
				delete[] c;
			}
			break;
		case 2:
			break;
	}
}

void Qt_TCP_ClientServerApp::clientDisconnect(void)
{
	AddMessage("Client disconnected...");
	SetButtons(true);
	NetStatus=NET_RESET;
	ReleaseIOBuffer();
}

void Qt_TCP_ClientServerApp::clientError(QAbstractSocket::SocketError socketError)
{
	switch (socketError)
	{
		case QAbstractSocket::RemoteHostClosedError:
			AddMessage("Remote Host Closed error: "+clientSocket->errorString());
			on_stopButton_clicked();
			break;
		case QAbstractSocket::HostNotFoundError:
			AddMessage("Remote Host Not Found error: "+clientSocket->errorString());
			on_stopButton_clicked();
			break;
		case QAbstractSocket::ConnectionRefusedError:
			AddMessage("Remote Host Connection Refused error: "+clientSocket->errorString());
			on_stopButton_clicked();
			break;
		default:
			AddMessage("Client socket error: "+clientSocket->errorString());
			break;
	}
}

void Qt_TCP_ClientServerApp::udpRead(void)
{
	qint64 nbytes;

	switch (ui->recMode->currentIndex())
	{
		case 0:
		case 1:
			break;
		case 2:
			if (udpSocket->hasPendingDatagrams())
			{
				nbytes=udpSocket->pendingDatagramSize();
				if (nbytes>0)
				{
					if (nbytes!=UDP_DATA_SIZE)
					{
						QString s;
						QTextStream ts(&s);
						ts << "Incorrect UDP input size: " << nbytes;
						AddMessage(s);
					}

					char *c=new char[nbytes];

					qint64 nb=udpSocket->readDatagram(c,nbytes);

					if (c[0]!=UDPCheckSig)
					{
						UDPSigErrors++;
						UDPCheckSig=c[0];
					}
					UDPCheckSig++;
					if (UDPCheckSig>'Z') UDPCheckSig='A';
					NRec+=nb;

					delete[] c;
				}
			}
			break;
	}
}

void Qt_TCP_ClientServerApp::udpError(QAbstractSocket::SocketError socketError)
{
	switch (socketError)
	{
		case QAbstractSocket::RemoteHostClosedError:
			AddMessage("Remote Host Closed error: "+udpSocket->errorString());
			on_stopButton_clicked();
			break;
		case QAbstractSocket::HostNotFoundError:
			AddMessage("Remote Host Not Found error: "+udpSocket->errorString());
			on_stopButton_clicked();
			break;
		case QAbstractSocket::ConnectionRefusedError:
			AddMessage("Remote Host Connection Refused error: "+udpSocket->errorString());
			on_stopButton_clicked();
			break;
		default:
			AddMessage("Client socket error: "+udpSocket->errorString());
			break;
	}
}
